"""
Context
-------
As in the experiments before, the set of EC numbers predicted by our approach is to be compared with the set predicted by the approach of Oh et al. (2007).

Question
--------
Does the consensus/majority graph approach to core metabolism yield a similar set of EC numbers as the approach of Oh et al. (2007)?

Method
------
- extract EC numbers from Oh et al. (2007) by hand
- remove EC numbers with wildcards
-     get group of organisms 'Bacillus subtilis'
-     REPEAT for varying majority-percentages:
-         calculate EC numbers occuring in group's core metabolism
-         remove EC numbers with wildcards
-         overlap Oh's set with ours and print amount of EC numbers inside the intersection and falling off either side

Result
------

::

    Maj. %    others    both    ours
    100%:    135    313    235
     90%:    114    334    261
     80%:    108    340    267
     70%:    108    340    268
     60%:    107    341    268
     50%:    107    341    268
     40%:    107    341    270
     30%:    107    341    270
     20%:    107    341    271
     10%:    107    341    274
      1%:    106    342    283

Conclusion
----------
Even at only 1% (effectively 1 organism) majority, there are many EC numbers predicted by Oh et al., but not occuring anywhere in our model organisms.
This may be caused by:
1. EC number is associated with the organism, but not listed in one of KEGG's hand-drawn pathways. For example 1.13.11.24 is associated with all our 15 organisms, but not present in any pathway.
2. As seen in experiment :mod:`30`, there may be EC numbers predicted by Oh et al. which are outdated.
3. Oh et al. used a compilation of several sources, some may have predicted EC numbers for B. subtilis which never made their way into KEGG at all, which is our only source.


"""
from FEV_KEGG.Graph.Elements import EcNumber
from FEV_KEGG.Evolution.Taxonomy import NCBI
from FEV_KEGG.KEGG import Organism


if __name__ == '__main__':
    
    output = ['Maj. %\tothers\tboth\tours']
    
    #- extract EC numbers from Oh et al. (2007) by hand
    theirECnumberStrings = ['5.3.1.23', '1.4.1.13', '6.3.5.5', '6.3.1.2', '2.3.1.1', '3.5.1.47', '1.4.1.1', '1.5.1.2', '2.1.3.3', '2.7.2.8', '2.6.1.11', '3.5.3.11', '1.2.1.38', '2.6.1.2', '4.1.1.19', '1.2.1.11', '6.3.5.4', '2.7.2.4', '2.6.1.1', '4.1.1.20', '5.1.1.7', '1.3.1.26', '4.2.1.52', '1.1.1.3', '2.7.1.39', '4.2.3.1', '1.2.1.41', '2.7.2.11', '4.1.3.18', '1.2.1.27', '1.4.1.2', '4.1.3.27', '2.4.2.18', '5.4.99.5', '4.2.3.5', '4.1.2.15', '4.2.1.10', '4.2.3.4', '4.1.1.48', '1.3.1.12', '4.2.1.51', '5.3.1.24', '2.5.1.19', '1.1.1.25', '2.7.1.71', '4.2.1.20', '2.6.1.5', '4.2.1.22', '4.1.99.1', '2.1.2.1', '2.7.1.25', '4.2.99.8', '4.4.1.1', '4.4.1.8', '1.1.1.95', '2.6.1.52', '2.3.1.30', '2.6.1.9', '2.6.1.58', '1.4.3.3', '2.5.1.6', '2.6.1.16', '2.1.1.13', '2.3.2.2', '1.5.3.1', '3.2.2.16', '4.1.1.50', '2.7.1.100', '2.5.1.16', '2.5.1.22', '2.6.1.19', '3.5.1.16', '4.2.99.8', '4.2.99.8', '5.1.1.1', '3.5.3.1', '4.3.2.1', '6.3.4.5', '2.4.2.17', '4.2.1.9', '4.2.1.9', '3.5.3.8', '3.5.1.31', '3.5.1.2', '5.1.1.3', '2.3.1.29', '1.1.1.31', '4.3.1.3', '1.1.1.23', '3.1.3.15', '1.5.1.12', '2.3.1.46', '4.2.1.19', '1.4.1.9', '2.6.1.42', '1.1.1.85', '4.2.1.33', '4.1.3.12', '3.5.2.7', '1.1.1.86', '1.1.1.86', '2.6.1.42', '1.4.1.9', '5.4.3.2', '4.1.1.18', '1.2.1.27', '1.2.1.27', '2.6.1.13', '2.3.1.35', '1.5.1.12', '3.5.4.19', '3.6.1.31', '5.3.1.16', '1.5.1.2', '3.1.3.3', '3.5.1.18', '4.3.1.18', '4.3.1.17', '4.2.99.9', '4.2.99.9', '4.2.99.9', '1.2.1.16', '1.1.1.103', '4.2.1.49', '3.5.1.5', '1.4.1.9', '2.6.1.42', '3.5.1.1', '4.3.1.1', '2.3.1.57', '2.3.1.57', '1.2.1.27', '2.1.1.37', '3.1.2.4', '3.1.3.3', '4.1.1.29', '4.2.99.9', '1.13.11.24', '1.1.1.2', '1.1.1.4', '4.1.1.5', '1.1.1.94', '2.7.1.12', '2.7.1.15', '1.1.1.14', '6.2.1.1', '6.4.1.1', '1.2.1.12', '1.1.1.44', '5.4.2.2', '5.1.3.1', '5.3.1.6', '1.1.99.5', '1.1.1.49', '2.7.2.1', '1.2.1.12', '4.2.1.11', '4.1.2.13', '1.1.1.17', '4.1.1.49', '1.1.1.40', '1.1.1.21', '3.2.1.108', '5.1.3.14', '4.2.1.42', '4.2.1.40', '4.1.1.2', '1.1.3.15', '5.4.2.10', '3.5.1.25', '1.1.1.267', '3.5.99.6', '3.2.1.55', '4.1.2.13', '4.2.1.7', '5.3.1.12', '1.1.1.57', '4.2.1.8', '1.1.1.58', '5.3.1.12', '2.7.1.2', '5.3.1.17', '1.1.1.127', '2.7.1.45', '4.1.2.14', '3.2.1.20', '3.2.1.48', '2.7.1.15', '1.1.1.21', '4.1.1.47', '1.1.1.60', '1.1.1.83', '4.1.1.73', '1.1.1.93', '1.1.1.93', '1.2.1.3', '5.1.3.2', '1.3.99.1', '6.2.1.5', '4.1.3.31', '4.1.3.30', '4.2.1.3', '4.1.3.7', '4.2.1.2', '1.1.1.42', '1.1.1.37', '3.2.1.21', '3.2.1.21', '3.2.1.26', '3.2.1.1', '3.2.1.86', '1.2.1.19', '3.6.1.13', '1.2.1.46', '1.2.1.3', '1.2.1.3', '1.2.1.3', '5.3.1.4', '3.2.1.21', '2.7.2.7', '4.2.1.41', '2.7.1.56', '3.1.3.11', '2.7.1.4', '1.1.1.118', '2.7.1.41', '5.3.1.9', '2.7.1.6', '3.2.1.22', '3.2.1.22', '3.2.1.22', '3.2.1.22', '2.7.7.10', '2.4.1.18', '1.2.1.21', '2.4.1.1', '2.4.1.21', '2.7.7.27', '2.7.1.30', '4.1.3.5', '3.2.1.26', '3.2.1.122', '2.3.1.79', '5.3.1.8', '4.2.3.3', '2.8.3.5', '2.3.1.19', '2.7.1.11', '3.2.1.86', '3.1.3.18', '5.4.2.6', '3.1.3.8', '5.4.2.8', '6.4.1.3', '2.7.9.2', '2.7.1.47', '2.7.1.16', '5.1.3.4', '2.7.1.5', '3.2.1.86', '2.2.1.2', '2.2.1.1', '2.2.1.1', '3.2.1.93', '5.1.3.14', '5.3.1.5', '5.3.1.5', '2.7.1.17', '2.3.1.8', '1.2.1.51', '2.7.1.11', '5.3.1.9', '5.4.2.1', '2.7.1.40', '5.3.1.1', '2.7.2.3', '5.3.1.14', '1.1.1.18', '4.1.3.19', '6.2.1.1', '5.3.1.3', '4.3.1.7', '3.5.1.49', '4.1.2.20', '2.7.1.31', '5.4.99.11', '1.2.1.22', '3.1.3.25', '2.4.1.8', '3.1.1.31', '4.1.2.19', '1.1.99.21', '4.1.2.40', '5.1.3.7', '3.5.3.19', '3.2.1.37', '2.7.8.13', '2.7.7.9', '2.7.7.39', '1.1.1.22', '2.7.7.23', '6.3.2.4', '3.5.1.28', '3.2.1.52', '6.3.2.13', '2.5.1.7', '6.3.2.9', '6.3.2.8', '1.1.1.158', '6.3.2.15', '2.3.1.157', '3.6.1.27', '1.5.1.5', '1.7.99.5', '2.7.7.1', '2.7.7.3', '2.7.1.24', '1.1.1.169', '2.5.1.30', '5.4.99.6', '2.1.2.11', '2.4.2.11', '3.5.1.19', '2.7.7.18', '4.1.3.36', '6.3.2.1', '2.7.1.34', '6.2.1.26', '2.7.7.58', '2.3.1.47', '1.1.1.193', '1.5.1.3', '6.3.2.12', '2.5.1.15', '2.5.1.9', '3.5.4.16', '3.5.4.9', '4.1.2.25', '2.6.1.62', '2.8.1.6', '6.2.1.14', '6.3.3.3', '3.5.4.26', '2.7.1.50', '2.7.6.3', '2.7.4.7', '2.5.1.9', '2.7.4.16', '2.5.1.3', '4.1.1.11', '1.4.3.16', '1.3.1.28', '1.5.1.3', '2.7.7.2', '3.5.1.10', '3.5.4.25', '3.3.2.1', '4.1.1.71', '2.7.1.33', '1.3.3.4', '4.2.3.12', '2.7.1.26', '2.4.2.19', '6.3.1.5', '3.5.99.2', '1.3.3.3', '4.99.1.1', '5.4.3.8', '4.3.1.8', '2.1.1.107', '4.2.1.24', '4.2.1.75', '4.1.1.37', '2.7.1.49', '3.1.3.5', '4.1.1.36', '6.3.2.5', '2.3.1.15', '3.1.3.27', '2.7.7.41', '4.1.1.65', '6.4.1.2', '1.2.1.25', '1.2.1.25', '1.2.1.25', '2.5.1.1', '5.3.3.2', '2.3.1.41', '2.3.1.41', '2.3.1.41', '2.3.1.41', '2.3.1.41', '2.3.1.41', '2.3.1.41', '2.3.1.41', '2.3.1.41', '2.3.1.41', '2.3.1.41', '2.3.1.41', '2.3.1.41', '2.3.1.41', '2.7.8.7', '6.2.1.3', '6.2.1.3', '6.2.1.3', '6.2.1.3', '6.2.1.3', '6.2.1.3', '6.2.1.3', '6.2.1.3', '6.2.1.3', '6.2.1.3', '6.2.1.3', '6.2.1.3', '6.2.1.3', '2.5.1.10', '2.7.8.8', '2.3.1.51', '2.3.1.16', '2.3.1.9', '2.3.1.16', '2.3.1.16', '2.3.1.16', '2.3.1.16', '2.3.1.16', '2.3.1.16', '1.3.99.2', '1.3.99.2', '1.3.99.3', '1.3.99.2', '1.3.99.3', '1.3.99.3', '1.3.99.3', '1.3.99.3', '1.3.99.3', '3.1.4.14', '2.7.1.107', '4.2.1.17', '4.2.1.17', '4.2.1.17', '4.2.1.17', '4.2.1.17', '4.2.1.17', '4.2.1.17', '4.2.1.17', '4.2.1.17', '6.2.1.3', '3.1.4.46', '3.1.4.46', '3.1.4.46', '1.1.1.35', '1.1.1.35', '1.1.1.35', '1.1.1.35', '1.1.1.35', '1.1.1.35', '1.1.1.35', '1.1.1.35', '1.1.1.35', '4.1.3.4', '2.5.1.31', '2.5.1.29', '2.5.1.33', '1.1.1.27', '1.6.5.3', '1.7.99.4', '3.6.3.14', '1.2.1.2', '1.5.1.30', '1.8.1.9', '1.9.3.1', '1.10.2.2', '1.5.1.29', '1.1.1.27', '3.1.3.5', '3.1.3.5', '3.1.3.5', '3.1.3.5', '3.1.3.5', '3.1.3.5', '3.1.3.5', '3.1.3.5', '3.1.3.5', '3.1.3.5', '3.1.4.16', '3.1.4.16', '3.1.4.16', '3.1.4.16', '3.1.3.6', '3.1.3.6', '3.1.3.6', '3.1.3.6', '2.7.1.23', '2.7.1.23', '2.7.1.23', '2.7.1.23', '2.7.1.23', '2.7.1.23', '3.2.2.9', '3.5.4.12', '3.5.4.12', '1.17.4.1', '1.17.4.1', '1.17.4.1', '1.17.4.1', '2.7.6.1', '3.5.4.3', '3.5.4.5', '2.4.2.2', '5.4.2.7', '5.4.2.7', '2.4.2.4', '2.4.2.2', '3.5.4.14', '2.7.1.74', '3.5.4.2', '2.7.1.20', '3.5.2.5', '2.7.4.14', '2.7.4.14', '2.7.4.11', '2.7.1.76', '2.7.1.113', '4.1.2.4', '2.7.4.9', '3.6.1.23', '2.7.4.8', '1.7.1.7', '2.7.6.5', '2.4.2.8', '2.7.4.6', '2.7.4.6', '2.7.4.6', '2.7.4.6', '2.7.4.6', '2.7.4.6', '2.7.4.6', '6.3.2.6', '2.4.2.1', '2.4.2.1', '2.4.2.1', '2.4.2.1', '2.4.2.1', '2.4.2.1', '2.4.2.1', '2.4.2.1', '2.4.2.2', '2.4.2.1', '2.7.1.21', '2.1.1.45', '1.7.3.3', '2.7.1.48', '2.7.1.48', '2.7.1.48', '1.1.1.204', '2.4.2.22', '6.3.4.13', '6.3.3.1', '6.3.5.3', '6.3.4.2', '2.7.4.3', '2.4.2.7', '4.3.2.2', '4.3.2.2', '6.3.4.4', '2.1.2.3', '2.1.3.2', '6.3.4.2', '1.3.3.1', '3.5.2.3', '2.1.2.2', '2.7.4.8', '2.4.2.14', '6.3.5.2', '3.5.4.10', '1.1.1.205', '2.7.4.6', '2.4.2.10', '2.4.2.9', '2.7.4.3', '3.5.3.4', '3.5.4.1', '2.7.4.6', '4.1.1.23', '2.7.4.4', '1.1.1.204', '1.2.1.8', '1.1.99.1', '3.5.2.6', '2.1.2.9', '1.11.1.9', '6.1.1.10', '3.5.1.11', '1.15.1.1', '6.1.1.17', '1.11.1.6', '3.6.1.15', '4.2.1.1', '2.7.7.4', '1.8.1.2', '3.1.3.1', '3.6.1.1', '3.1.3.1', '3.6.1.2']
    theirECnumbers = set()
    
    for string in theirECnumberStrings:
        theirECnumbers.add( EcNumber(string) )
    
    #- remove EC numbers with wildcards
    theirECnumbers = EcNumber.removeWildcards(theirECnumbers)
    
    taxonomy = NCBI.getTaxonomy()
            
    #-     get group of organisms 'Bacillus subtilis'
    organisms = taxonomy.getOrganismAbbreviationsByPath('Bacillus subtilis')
    group = Organism.Group( organisms )
    
    #-     REPEAT for varying majority-percentages:
    for percentage in [100, 90, 80, 70, 60, 50, 40, 30, 20, 10 , 1]:
    
        #-         calculate EC numbers occuring in group's core metabolism
        ourECnumbers = group.majorityEcGraph(majorityPercentage = percentage, noMultifunctional = False).getECs()
        
        #-         remove EC numbers with wildcards
        ourECnumbers = EcNumber.removeWildcards(ourECnumbers)
        
        #-         overlap Oh's set with ours and print amount of EC numbers inside the intersection and falling off either side
        onlyInTheirs = theirECnumbers.difference( ourECnumbers ) 
        inBoth = theirECnumbers.intersection( ourECnumbers )
        onlyInOurs = ourECnumbers.difference( theirECnumbers )
        
        output.append(str(percentage) + '%:\t' + str(len(onlyInTheirs)) + '\t' + str(len(inBoth)) + '\t' + str(len(onlyInOurs)) )
        
    for line in output:
        print(line)